@@ -9,8 +9,9 @@ use IO::Socket::INET;
 use IO::Select;
 use Net::WebSocket::Server::Connection;
 use Time::HiRes qw(time);
+use List::Util qw(min);
 
-our $VERSION = '0.002003';
+our $VERSION = '0.003000';
 $VERSION = eval $VERSION;
 
 sub new {
@@ -19,21 +20,96 @@ sub new {
   my %params = @_;
 
   my $self = {
-    listen      => 80,
-    silence_max => 20,
-    on_connect  => sub{},
+    listen         => 80,
+    silence_max    => 20,
+    tick_period    => 0,
+    watch_readable => [],
+    watch_writable => [],
+    on_connect     => sub{},
+    on_tick        => sub{},
+    on_shutdown    => sub{},
   };
 
   while (my ($key, $value) = each %params ) {
     croak "Invalid $class parameter '$key'" unless exists $self->{$key};
-    croak "$class parameter '$key' expects a coderef" if ref $self->{$key} eq 'CODE' && ref $value ne 'CODE';
+    croak "$class parameter '$key' expected type is ".ref($self->{$key}) if ref $self->{$key} && ref $value ne ref $self->{$key};
     $self->{$key} = $value;
   }
 
+  bless $self, $class;
+
   # send a ping every silence_max by checking whether data was received in the last silence_max/2
   $self->{silence_checkinterval} = $self->{silence_max} / 2;
 
-  bless $self, $class;
+  foreach my $watchtype (qw(readable writable)) {
+    $self->{"select_$watchtype"} = IO::Select->new();
+    my $key = "watch_$watchtype";
+    croak "$class parameter '$key' expects an arrayref containing an even number of elements" unless @{$self->{$key}} % 2 == 0;
+    my @watch = @{$self->{$key}};
+    $self->{$key} = {};
+    $self->_watch($watchtype, @watch);
+  }
+
+  return $self;
+}
+
+sub watch_readable {
+  my $self = shift;
+  croak "watch_readable expects an even number of arguments" unless @_ % 2 == 0;
+  $self->_watch(readable => @_);
+}
+
+sub watched_readable {
+  my $self = shift;
+  return $self->{watch_readable}{$_[0]}{cb} if @_;
+  return map {$_->{fh}, $_->{cb}} values %{$self->{watch_readable}};
+}
+
+sub watch_writable {
+  my $self = shift;
+  croak "watch_writable expects an even number of arguments" unless @_ % 2 == 0;
+  $self->_watch(writable => @_);
+}
+
+sub watched_writable {
+  my $self = shift;
+  return $self->{watch_writable}{$_[0]}{cb} if @_;
+  return map {$_->{fh}, $_->{cb}} values %{$self->{watch_writable}};
+}
+
+sub _watch {
+  my $self = shift;
+  my $watchtype = shift;
+  croak "watch_$watchtype expects an even number of arguments after the type" unless @_ % 2 == 0;
+  for (my $i = 0; $i < @_; $i+=2) {
+    my ($fh, $cb) = ($_[$i], $_[$i+1]);
+    croak "watch_$watchtype expects the second value of each pair to be a coderef, but element $i was not" unless ref $cb eq 'CODE';
+    if ($self->{"watch_$watchtype"}{$fh}) {
+      carp "watch_$watchtype was given a filehandle at index $i which is already being watched; ignoring!";
+      next;
+    }
+    $self->{"select_$watchtype"}->add($fh);
+    $self->{"watch_$watchtype"}{$fh} = {fh=>$fh, cb=>$cb};
+  }
+}
+
+sub unwatch_readable {
+  my $self = shift;
+  $self->_unwatch(readable => @_);
+}
+
+sub unwatch_writable {
+  my $self = shift;
+  $self->_unwatch(writable => @_);
+}
+
+sub _unwatch {
+  my $self = shift;
+  my $watchtype = shift;
+  foreach my $fh (@_) {
+    $self->{"select_$watchtype"}->remove($fh);
+    delete $self->{"watch_$watchtype"}{$fh};
+  }
 }
 
 sub on {
@@ -58,37 +134,61 @@ sub start {
     ReuseAddr => 1,
   ) || croak "failed to listen on port $self->{listen}: $!" unless ref $self->{listen};
 
-  $self->{select} = IO::Select->new($self->{listen});
-  $self->{conns} = {};
-  $self->{silence_nextcheck} = $self->{silence_max} ? (time + $self->{silence_checkinterval}) : 0;
+  $self->{select_readable}->add($self->{listen});
 
-  while ($self->{select}->count) {
-    my $silence_checktimeout = $self->{silence_max} ? ($self->{silence_nextcheck} - time) : undef;
-    my @ready = $self->{select}->can_read($silence_checktimeout);
-    foreach my $fh (@ready) {
+  $self->{conns} = {};
+  my $silence_nextcheck = $self->{silence_max} ? (time + $self->{silence_checkinterval}) : 0;
+  my $tick_next = $self->{tick_period} ? (time + $self->{tick_period}) : 0;
+
+  while (%{$self->{conns}} || $self->{listen}->opened) {
+    my $silence_checktimeout = $self->{silence_max} ? ($silence_nextcheck - time) : undef;
+    my $tick_timeout = $self->{tick_period} ? ($tick_next - time) : undef;
+    my $timeout = min(grep {defined} ($silence_checktimeout, $tick_timeout));
+
+    my ($ready_read, $ready_write, undef) = IO::Select->select($self->{select_readable}, $self->{select_writable}, undef, $timeout);
+    foreach my $fh ($ready_read ? @$ready_read : ()) {
       if ($fh == $self->{listen}) {
         my $sock = $self->{listen}->accept;
         next unless $sock;
         my $conn = new Net::WebSocket::Server::Connection(socket => $sock, server => $self);
         $self->{conns}{$sock} = {conn=>$conn, lastrecv=>time};
-        $self->{select}->add($sock);
+        $self->{select_readable}->add($sock);
         $self->{on_connect}($self, $conn);
-      } else {
+      } elsif ($self->{watch_readable}{$fh}) {
+        $self->{watch_readable}{$fh}{cb}($self, $fh);
+      } elsif ($self->{conns}{$fh}) {
         my $connmeta = $self->{conns}{$fh};
         $connmeta->{lastrecv} = time;
         $connmeta->{conn}->recv();
+      } else {
+        warn "filehandle $fh became readable, but no handler took responsibility for it; removing it";
+        $self->{select_readable}->remove($fh);
+      }
+    }
+
+    foreach my $fh ($ready_write ? @$ready_write : ()) {
+      if ($self->{watch_writable}{$fh}) {
+        $self->{watch_writable}{$fh}{cb}($self, $fh);
+      } else {
+        warn "filehandle $fh became writable, but no handler took responsibility for it; removing it";
+        $self->{select_writable}->remove($fh);
       }
     }
 
     if ($self->{silence_max}) {
       my $now = time;
-      if ($self->{silence_nextcheck} < $now) {
-        my $lastcheck = $self->{silence_nextcheck} - $self->{silence_checkinterval};
+      if ($silence_nextcheck < $now) {
+        my $lastcheck = $silence_nextcheck - $self->{silence_checkinterval};
         $_->{conn}->send('ping') for grep { $_->{lastrecv} < $lastcheck } values %{$self->{conns}};
 
-        $self->{silence_nextcheck} = $now + $self->{silence_checkinterval};
+        $silence_nextcheck = $now + $self->{silence_checkinterval};
       }
     }
+
+    if ($self->{tick_period} && $tick_next < time) {
+      $self->{on_tick}($self);
+      $tick_next += $self->{tick_period};
+    }
   }
 }
 
@@ -96,14 +196,15 @@ sub connections { map {$_->{conn}} values %{$_[0]{conns}} }
 
 sub shutdown {
   my ($self) = @_;
-  $self->{select}->remove($self->{listen});
+  $self->{on_shutdown}($self);
+  $self->{select_readable}->remove($self->{listen});
   $self->{listen}->close();
   $_->disconnect(1001) for $self->connections;
 }
 
 sub disconnect {
   my ($self, $fh) = @_;
-  $self->{select}->remove($fh);
+  $self->{select_readable}->remove($fh);
   $fh->close();
   delete $self->{conns}{$fh};
 }
@@ -135,6 +236,19 @@ Simple echo server for C<utf8> messages.
         },
     )->start;
 
+Server that sends the current time to all clients every second.
+
+    use Net::WebSocket::Server;
+
+    my $ws = Net::WebSocket::Server->new(
+        listen => 8080,
+        tick_period => 1,
+        on_tick => sub {
+            my ($serv) = @_;
+            $_->send_utf8(time) for $serv->connections;
+        },
+    )->start;
+
 Broadcast-echo server for C<utf8> and C<binary> messages with origin testing.
 
     use Net::WebSocket::Server;
@@ -208,7 +322,7 @@ and speaks SSL, such as L<IO::Socket::SSL|IO::Socket::SSL>.  For example:
     Net::WebSocket::Server->new(
         listen => $ssl_server,
         on_connect => sub { ... },
-    )
+    )->start;
 
 =item C<silence_max>
 
@@ -217,11 +331,36 @@ socket.  Every C<silence_max/2> seconds, each connection is checked for
 whether data was received since the last check; if not, a WebSocket ping
 message is sent.  Set to C<0> to disable.  Default C<20>.
 
+=item C<tick_period>
+
+The amount of time in seconds between C<tick> events.  Set to C<0> to disable.
+Default C<0>.
+
 =item C<on_C<$event>>
 
 The callback to invoke when the given C<$event> occurs, such as C<on_connect>.
 See L</EVENTS>.
 
+=item C<watch_readable>
+
+=item C<watch_writable>
+
+Each of these takes an I<arrayref> of C<< $filehandle => $callback >> pairs to be
+passed to the corresponding method.  Default C<[]>.  See
+L<watch_readable()|/watch_readable(@pairs)> and
+L<watch_writable()|/watch_writable(@pairs)>.  For example:
+
+    Net::WebSocket::Server->new(
+        # ...other relevant arguments...
+        watch_readable => [
+            \*STDIN => \&on_stdin,
+        ],
+        watch_writable => [
+            $log1_fh => sub { ... },
+            $log2_fh => sub { ... },
+        ],
+    )->start;
+
 =back
 
 =back
@@ -234,7 +373,7 @@ See L</EVENTS>.
 
     $server->on(
         connect => sub { ... },
-    )
+    );
 
 Takes a list of C<< $event => $callback >> pairs; C<$event> names should not
 include an C<on_> prefix.  Typically, events are configured once via the
@@ -264,6 +403,44 @@ function anyway.
 Closes the listening socket and cleanly disconnects all clients, causing the
 L<start()|/start> method to return.
 
+=item C<watch_readable(I<@pairs>)>
+
+    $server->watch_readable(
+      \*STDIN => \&on_stdin,
+    );
+
+Takes a list of C<< $filehandle => $callback >> pairs.  The given filehandles
+will be monitored for readability; when readable, the given callback will be
+invoked.  Arguments passed to the callback are the server itself and the
+filehandle which became readable.
+
+=item C<watch_writable(I<@pairs>)>
+
+    $server->watch_writable(
+      $log1_fh => sub { ... },
+      $log2_fh => sub { ... },
+    );
+
+Takes a list of C<< $filehandle => $callback >> pairs.  The given filehandles
+will be monitored for writability; when writable, the given callback will be
+invoked.  Arguments passed to the callback are the server itself and the
+filehandle which became writable.
+
+=item C<watched_readable([I<$filehandle>])>
+
+=item C<watched_writable([I<$filehandle>])>
+
+These methods return a list of C<< $filehandle => $callback >> pairs that are
+curently being watched for readability / writability.  If a filehandle is
+given, its callback is returned, or C<undef> if it isn't being watched.
+
+=item C<unwatch_readable(I<@filehandles>)>
+
+=item C<unwatch_writable(I<@filehandles>)>
+
+These methods cause the given filehandles to no longer be watched for
+readability / writability.
+
 =back
 
 =head1 EVENTS
@@ -282,6 +459,20 @@ L<Net::WebSocket::Server::Connection|Net::WebSocket::Server::Connection>
 object.  Arguments passed to the callback are the server accepting the
 connection and the new connection object itself.
 
+=item C<tick(I<$server>)>
+
+Invoked every L<tick_period|/tick_period> seconds, or never if
+L<tick_period|/tick_period> is C<0>.  Useful to perform actions that aren't in
+response to a message from a client.  Arguments passed to the callback are only
+the server itself.
+
+=item C<shutdown(I<$server>)>
+
+Invoked immediately before the server shuts down due to the L<shutdown()>
+method being invoked.  Any client connections will still be available until
+the event handler returns.  Arguments passed to the callback are only the
+server that is being shut down.
+
 =back
 
 =head1 AUTHOR
